{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from selenium import webdriver\n",
    "from selenium.webdriver.common.by import By\n",
    "from selenium.webdriver.common.keys import Keys\n",
    "import csv\n",
    "import os\n",
    "import re"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "def open_browser():\n",
    "    chrome_options = webdriver.ChromeOptions()\n",
    "    chrome_options.add_argument('--remote-debugging-port=9222')\n",
    "    chrome_options.add_argument('webdriver.chrome.driver=chromedriver/chromedriver')\n",
    "    driver = webdriver.Chrome(options=chrome_options)\n",
    "    return driver\n",
    "\n",
    "def manual_login(base_url):\n",
    "    url = os.path.join(base_url, 'index.jsp')\n",
    "    driver = open_browser()\n",
    "    driver.get(url)\n",
    "    ### user now logs in. After logging in, enter 'y' to continue.\n",
    "    ok = input('Ready? ([y]/n): ')\n",
    "    if ok.lower() == 'n':\n",
    "        exit()\n",
    "    url = os.path.join(base_url, 'menu.jsp')\n",
    "    return driver, url"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Scraper:\n",
    "\n",
    "\t# relevant ids for the scraper\n",
    "    # note that these can change as and when updates are made to asc\n",
    "    ids = {\n",
    "        'Academic':             ('//*[@id=\"ygtvlabelel1\"]',0),\n",
    "        'All About Courses':    ('//*[@id=\"ygtvlabelel7\"]',0),\n",
    "        'Running Courses':      ('//a[@id=\"4_59_1\"]',1),\n",
    "        'Grading Statistics':   ('//a[@id=\"4_71_1\"]',1),\n",
    "    }\n",
    "        \n",
    "    possible_grades = ['AA', 'AB', 'AP', 'AU', 'BB', 'BC', 'CC', 'CD', 'DD', 'FF', 'FR', 'PP', 'NP', 'DX', 'W', 'II']\n",
    "\n",
    "    def __init__(self, driver: webdriver.Chrome, url: str) -> None:\n",
    "        self.url = url\n",
    "        self.driver = driver\n",
    "        self.driver.get(self.url)\n",
    "        self.window_idx = 0\n",
    "\n",
    "    def refresh(self):\n",
    "        self.switch_window(0)\n",
    "        self.driver.refresh()\n",
    "\n",
    "    def switch_window(self, window_idx: int = 0):\n",
    "        self.window_idx = window_idx\n",
    "        self.driver.switch_to.window(self.driver.window_handles[self.window_idx])\n",
    "\n",
    "    def click_path(self, path: str):\n",
    "        path = path.split('/')\n",
    "        print(path)\n",
    "        for p in path:\n",
    "            xpath, window_inc = self.ids[p]\n",
    "            self.driver.find_element(By.XPATH, value=xpath).click()\n",
    "            self.window_idx += window_inc\n",
    "            if window_inc:\n",
    "                self.driver.switch_to.window(self.driver.window_handles[self.window_idx])\n",
    "        return self.window_idx\n",
    "    \n",
    "    def get_all_depts(self, year: str, semester: str):\n",
    "        self.choose_option('year', year)\n",
    "        self.choose_option('semester', semester)\n",
    "        self.driver.find_element(By.XPATH, value=\"//input[@name='submit']\").click()\n",
    "        all_depts = self.driver.find_elements(By.XPATH, value=\"//a[contains(@href, 'deptcd')]\")\n",
    "        depts = [re.search(r'(?<=deptcd=)[A-Z,]+', dept.get_attribute('href')).group() for dept in all_depts]\n",
    "        depts = [dept.replace(',', '-') for dept in depts]\n",
    "        # depts = '-'.join(depts).split('-')\n",
    "        return depts\n",
    "    \n",
    "    def get_courses(self, year: str, semester: str, dept: str, savefilename: str):\n",
    "        '''gets all courses offered by {dept} in {year}-{semester}, saves to {savefilename}'''\n",
    "        self.choose_option('year', year)\n",
    "        self.choose_option('semester', semester)\n",
    "        self.driver.find_element(By.XPATH, value=\"//input[@name='submit']\").click()\n",
    "        self.driver.find_element(By.XPATH, value=f'//a[contains(@href, \"{dept}\")]').click()\n",
    "        all_rows = self.driver.find_elements(By.XPATH, value='//table[1]/tbody/tr')[4:]\n",
    "        courses = []\n",
    "        n = len(all_rows)\n",
    "        for i in range(n):\n",
    "            all_rows = self.driver.find_elements(By.XPATH, value='//table[1]/tbody/tr')[4:] # refresh\n",
    "            row = all_rows[i]\n",
    "            cols = row.find_elements(By.XPATH, value='td')\n",
    "            if len(cols) != 14: continue # ignore spurious rows (see html to figure out what this is for)\n",
    "            code = cols[2].text\n",
    "            name = cols[3].text\n",
    "            instructor = cols[6].text.split('- ')[1:]\n",
    "            for i, ins in enumerate(instructor):\n",
    "                instructor[i] = ins.split('\\n')[0]\n",
    "            instructor = '\\n'.join(instructor)\n",
    "            slot = cols[8].text.split('\\n')[0]\n",
    "\n",
    "            # get credits, text, desc\n",
    "            cols[2].find_element(By.XPATH, value='a').click()\n",
    "            summary = self.driver.find_elements(By.XPATH, value='//table[1]/tbody/tr')\n",
    "            creds = summary[2].text.replace('Total Credits ', '')\n",
    "            refs = summary[9].text.replace('Text Reference ', '')\n",
    "            if refs == ' ':\n",
    "                refs = 'None'\n",
    "            summ = summary[10].text.replace('Description ', '')\n",
    "            if summ == ' ':\n",
    "                summ = 'None'\n",
    "            self.driver.back()\n",
    "            courses.append((code, name, instructor, slot, creds, refs, summ))\n",
    "        self.driver.back()\n",
    "\n",
    "        # save to csv\n",
    "        with open(savefilename, 'w') as f:\n",
    "            writer = csv.writer(f)\n",
    "            writer.writerow(['Code', 'Name', 'Instructor', 'Slot', 'Credits', 'Text Reference', 'Description'])\n",
    "            writer.writerows(courses)\n",
    "        return courses\n",
    "    \n",
    "    def choose_option(self, name: str, option: str):\n",
    "        '''select an option from a dropdown menu'''\n",
    "        select = self.driver.find_element(By.XPATH, f\"//select[@name='{name}']\")\n",
    "        select.click()\n",
    "        options = select.find_elements(By.XPATH, \"//option\")\n",
    "        for opt in options:\n",
    "            if opt.text == option:\n",
    "                opt.click()\n",
    "                break\n",
    "\n",
    "    def put(self, name: str, text: str):\n",
    "        '''fill in a text field named {name} with {text}'''\n",
    "        elem = self.driver.find_element(By.XPATH, f\"//input[@name='{name}']\")\n",
    "        elem.send_keys(Keys.COMMAND, 'a')\n",
    "        elem.send_keys(text)\n",
    "\n",
    "    def collect_grade(self, code):\n",
    "        self.put('txtcrsecode', code)\n",
    "        submit_button = self.driver.find_element(By.XPATH, \"//input[@name='submit']\")\n",
    "        submit_button.click()\n",
    "        tables = self.driver.find_elements(By.XPATH, '//table')[3:] # ignore first 3 tables\n",
    "        grades = []\n",
    "        for table in tables:\n",
    "            data = table.text.split('\\n')\n",
    "            sect = data[0].split(' ')[-3]\n",
    "            full_code = code + ('-' + sect if sect != 'for' else '')\n",
    "            data = list(map(lambda x: x.split(' '), data[1:-1]))\n",
    "            grade_dict = {gr: 0 for gr in self.possible_grades}\n",
    "            for grade, count in data:\n",
    "                assert grade in self.possible_grades, f'{grade} not in {self.possible_grades}'\n",
    "                grade_dict[grade] = int(count)\n",
    "            grades.append((full_code, grade_dict.copy()))\n",
    "        self.driver.back()\n",
    "        return grades\n",
    "\n",
    "    def get_grades(self, year: str, semester: str, courses: list[str], savefilename: str):\n",
    "        grades = []\n",
    "        self.choose_option('year', year)\n",
    "        self.choose_option('semester', semester)\n",
    "        grades = sum((self.collect_grade(course) for course in courses), [])\n",
    "        with open(savefilename, 'w') as f:\n",
    "            writer = csv.writer(f)\n",
    "            writer.writerow(['Code'] + self.possible_grades)\n",
    "            for code, grade_dict in grades:\n",
    "                writer.writerow([code] + list(grade_dict.values()))\n",
    "        return grades"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "year = '2022-2023'\n",
    "semester = '1 - Autumn'\n",
    "course_dir = f'courses-{year}-{semester}'\n",
    "grade_dir = f'grades-{year}-{semester}'\n",
    "os.makedirs(course_dir, exist_ok=True)\n",
    "os.makedirs(grade_dir, exist_ok=True)\n",
    "\n",
    "coursefilename = lambda dept: f'{course_dir}/{dept}.csv'\n",
    "gradefilename = lambda dept: f'{grade_dir}/{dept}.csv'"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "def courses_from_file(coursesfilename: str):\n",
    "    with open(coursesfilename, 'r') as f:\n",
    "        reader = csv.reader(f)\n",
    "        next(reader)\n",
    "        courses = sorted(set((map(lambda x: x[0], reader))))\n",
    "    return courses\n",
    "\n",
    "def get_all_courses(scraper: Scraper, depts: list[str], year: str, semester: str):\n",
    "    for dept in depts:\n",
    "        dept_short = dept.split('-')[0]\n",
    "        scraper.get_courses(year, semester, dept_short, coursefilename(dept))\n",
    "\n",
    "def get_all_grades(scraper: Scraper, depts: list[str], year: str, semester: str):\n",
    "    for dept in depts:\n",
    "        courses = courses_from_file(coursefilename(dept))\n",
    "        print(courses)\n",
    "        scraper.get_grades(year, semester, courses, gradefilename(dept))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "base_url = 'https://asc.iitb.ac.in/acadmenu'\n",
    "login = manual_login(base_url)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "scraper = Scraper(*login)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "scraper.click_path('Academic/All About Courses/Running Courses')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "depts = scraper.get_all_depts(year, semester)[:3]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "depts"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# in case of crashes, run this cell again\n",
    "depts = depts[len(os.listdir(course_dir)):]\n",
    "get_all_courses(scraper, depts, year, semester)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# computing grading stats"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "year_for_grade = year[:4]\n",
    "semester_for_grade = semester[0]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "scraper.switch_window(0)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "scraper.click_path('Grading Statistics')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "get_all_grades(scraper, depts, year_for_grade, semester_for_grade)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# some patches, mostly data cleaning\n",
    "courses_dir = f'courses-{year}-{semester}'\n",
    "cleaned_courses_dir = f'cleaned-courses-{year}-{semester}'\n",
    "\n",
    "def remove_duplicate_courses(courses_dir, cleaned_courses_dir):\n",
    "    for filename in os.listdir(courses_dir):\n",
    "        if not filename.endswith('.csv'):\n",
    "            continue\n",
    "        with open(f'{courses_dir}/{filename}', 'r') as f:\n",
    "            reader = csv.reader(f)\n",
    "            hdr = next(reader)\n",
    "            courses = {}\n",
    "            for row in reader:\n",
    "                if row[0] in courses:\n",
    "                    instrs = row[2].replace('\\n', ', ').split(', ')\n",
    "                    instrs = [instr for instr in instrs if instr != '']\n",
    "                    instrs = set(instrs)\n",
    "                    courses[row[0]][2] = courses[row[0]][2].union(instrs)\n",
    "                else:\n",
    "                    instrs = row[2].replace('\\n', ', ').split(', ')\n",
    "                    instrs = [instr for instr in instrs if instr != '']\n",
    "                    instrs = set(instrs)\n",
    "                    courses[row[0]] = row[:2] + [instrs] + row[3:]\n",
    "                if row[0] == 'MA 110':\n",
    "                    print(row[2])\n",
    "                    print(courses[row[0]])\n",
    "                    # exit()\n",
    "        with open(f'{cleaned_courses_dir}/{filename}', 'w') as f:\n",
    "            writer = csv.writer(f)\n",
    "            writer.writerow(hdr)\n",
    "            for code in sorted(courses):\n",
    "                writer.writerow(courses[code][:2] + [', '.join(courses[code][2])] + courses[code][3:])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "os.makedirs(cleaned_courses_dir, exist_ok=True)\n",
    "remove_duplicate_courses(courses_dir, cleaned_courses_dir)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "len(os.listdir('cleaned-' + course_dir)), len(os.listdir(grade_dir)), len(depts)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import statistics\n",
    "\n",
    "coursedir = f'cleaned-courses-{year}-{semester}'\n",
    "gradedir = f'grades-{year}-{semester}'\n",
    "possible_grades = ['AA', 'AB', 'AP', 'AU', 'BB', 'BC', 'CC', 'CD', 'DD', 'FF', 'FR', 'PP', 'NP', 'DX', 'W', 'II']\n",
    "grade_to_num = [10, 9, 10, 6, 8, 7, 6, 5, 4, 0, 0, 10, 0, 0, 0, 0]\n",
    "\n",
    "concat = open(f'allcourses-{year}-{semester}.csv', 'w')\n",
    "writer = csv.writer(concat)\n",
    "writer.writerow(['Code', 'Name', 'Instructor', 'Slot', 'Credits', 'Text Reference', 'Description', 'Mean', 'Median', 'Mode', 'Std. Dev.'] + possible_grades)\n",
    "\n",
    "depts = []\n",
    "for filename in sorted(os.listdir(coursedir)):\n",
    "    if filename.endswith('.csv'):\n",
    "        depts.append(filename[:-4])\n",
    "\n",
    "print(depts)\n",
    "\n",
    "grades = {}\n",
    "for dept in depts:\n",
    "    grades[dept] = {} # {code: [grades]}\n",
    "    with open(f'{gradedir}/{dept}.csv', 'r') as f:\n",
    "        reader = csv.reader(f)\n",
    "        next(reader)\n",
    "        for row in reader:\n",
    "            code = row[0]\n",
    "            grades[dept][code] = list(map(int, row[1:]))\n",
    "\n",
    "for dept in depts:\n",
    "    with open(f'{coursedir}/{dept}.csv', 'r') as f:\n",
    "        reader = csv.reader(f)\n",
    "        next(reader)\n",
    "        codes = set()\n",
    "        for row in reader:\n",
    "            code = row[0]\n",
    "            found = False\n",
    "            for code_sec in grades[dept]:\n",
    "                if code != code_sec.split('-')[0]:\n",
    "                    continue\n",
    "                found = True\n",
    "                grds = grades[dept][code_sec]\n",
    "                print(grds, dept, code)\n",
    "                grades_list = sum(([float(grade_points)]*num_given for grade_points, num_given in zip(grade_to_num, grds)), [])\n",
    "                # print(grades_list)\n",
    "                if len(grades_list) == 0:\n",
    "                    continue # pathos, man, pathos\n",
    "                if len(grades_list) > 1:\n",
    "                    mean = statistics.mean(grades_list)\n",
    "                    median = statistics.median(grades_list)\n",
    "                    mode = statistics.mode(grades_list)\n",
    "                    stddev = statistics.stdev(grades_list)\n",
    "                else:\n",
    "                    mean = median = mode = grades_list[0]\n",
    "                    stddev = 0.\n",
    "                writer.writerow([code_sec] + row[1:] + [mean, median, mode, stddev] + grds)\n",
    "            if not found:\n",
    "                grds = ['NA']*len(possible_grades)\n",
    "                mean = median = mode = stddev = 'NA'\n",
    "                writer.writerow(row + [mean, median, mode, stddev] + grds)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "env",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.14"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
